from typing import Callable, Optional, List, Dict
from fastapi import Query, Request
from .helper import parse_query_search, parse_query_sort, get_params_filter
from pydantic.types import Json
from pydantic import BaseModel
from .types import QuerySortDict
from .models import (
    AuthModel,
    QuerySortModel,
    PathParamModel,
    JoinOptions,
    JoinOptionModel
)


class GetQuerySearch:

    def __init__(
        self,
        option_filter: Optional[Dict] = None
    ):
        self.option_filter = option_filter

    def __call__(
        self,
        request: Request,
        search_spec: Optional[Json] = Query(None, alias="s"),
        filters: List[str] = Query(None, alias="filter"),
        ors: List[str] = Query(None, alias="or"),
    ):
        params_filter = getattr(request.state, "params_filter", None)
        auth_filter = getattr(request.state, "auth_filter", None)
        search = parse_query_search(
            search_spec=search_spec,
            ors=ors,
            filters=filters,
            option_filter=self.option_filter,
            auth_filter=auth_filter,
            params_filter=params_filter
        )
        return search


class GetQuerySorts:

    def __init__(self, option_sort: Optional[List[QuerySortDict]] = None):
        self.option_sort = option_sort
        if self.option_sort and isinstance(self.option_sort[0], BaseModel):
            self.option_sort = [item.model_dump() for item in self.option_sort]

    def __call__(
        self,
        sorts: List[str] = Query(None, alias="sort")
    ):
        if sorts:
            return parse_query_sort(sorts)
        if self.option_sort:
            return self.option_sort
        return []


class GetQueryJoins:

    def __init__(self, option_joins: Optional[JoinOptions] = None):
        self.option_joins = option_joins

    def __call__(
        self,
        loads: List[str] = Query(None, alias="load"),
        joins: List[str] = Query(None, alias="join"),
    ):

        join_options: JoinOptions = {
            **self.option_joins} if self.option_joins else {}
        if loads:
            for load_field in loads:
                if load_field in join_options:
                    join_options[load_field].select = True
                else:
                    join_options[load_field] = JoinOptionModel(
                        select=True, join=False)
        if joins:
            for join_field in joins:
                if join_field in join_options:
                    join_options[join_field].join = True
                else:
                    join_options[join_field] = JoinOptionModel(
                        select=False, join=True)
        return join_options


class GetQueryLoads:

    def __init__(self, option_joins: Optional[JoinOptions] = None):
        self.option_joins = option_joins

    def __call__(
        self,
        loads: List[str] = Query(None, alias="load"),
    ):

        join_options: JoinOptions = {
            **self.option_joins} if self.option_joins else {}
        if loads:
            for load_field in loads:
                if load_field in join_options:
                    join_options[load_field].select = True
                else:
                    join_options[load_field] = JoinOptionModel(
                        select=True, join=False)
        return join_options


class CrudAction():

    def __init__(self, feature: str, action_map: Dict, router_name: str):
        self.feature = feature
        self.action_map = action_map
        self.router_name = router_name

    def __call__(self, request: Request):
        request.state.feature = self.feature
        request.state.action = self.action_map.get(self.router_name).value


class StateAction():

    def __init__(self, auth: AuthModel, params: Dict[str, PathParamModel]):
        self.auth = auth
        self.params = params

    def __call__(self, request: Request):
        if self.auth:
            if self.auth.persist and isinstance(self.auth.persist, Callable):
                request.state.auth_persist = self.auth.persist(request)
            if self.auth.filter_ and isinstance(self.auth.filter_, Callable):
                request.state.auth_filter = self.auth.filter_(request)
        if self.params:
            params_filter = get_params_filter(self.params, request)
            request.state.params_filter = params_filter



from __future__ import annotations

import math
from typing import Generic, Sequence, TypeVar, Optional, Any
from fastapi import Query
from fastapi_pagination.bases import AbstractPage, AbstractParams, RawParams
from pydantic import BaseModel


T = TypeVar('T')


class Params(BaseModel, AbstractParams):
    page: Optional[int] = Query(default=None, gte=0, description='Page number')
    size: Optional[int] = Query(
        default=None,
        gte=0,
        le=100,
        description='Page size'
    )

    def to_raw_params(self) -> RawParams:
        return RawParams(
            limit=self.size,
            offset=self.size * (self.page - 1),
        )


class Page(AbstractPage[T], Generic[T]):
    items: Sequence[T]
    total: int
    page: int
    size: int
    pages: int
    __params_type__ = Params

    @classmethod
    def create(
        cls,
        items: Sequence[T],
        params: AbstractParams,
        *,
        total: Optional[int] = None,
        **kwargs: Any,
    ) -> Page[T]:
        size = params.size if params.size is not None else (total or None)
        page = params.page if params.page is not None else 1
        if size in {0, None}:
            pages = 0
        elif total is not None:
            pages = math.ceil(total / size)
        else:
            pages = None
        return cls(items=items, total=total, page=page, size=size, pages=pages)



import abc
from typing import Dict, List, Union, TypeVar, Generic, Optional
from fastapi import Request, BackgroundTasks
from fastapi_pagination.bases import AbstractPage
from ..types import QuerySortDict, ID_TYPE, CreateSchemaType, UpdateSchemaType
from ..models import JoinOptions

ModelType = TypeVar("ModelType")


class AbstractCrudService(Generic[ModelType], abc.ABC):

    @abc.abstractmethod
    async def crud_get_many(
        self,
        request: Optional[Request] = None,
        search: Optional[Dict] = None,
        include_deleted: Optional[bool] = False,
        soft_delete: Optional[bool] = False,
        sorts: List[QuerySortDict] = None,
        joins: Optional[JoinOptions] = None,
    ) -> Union[AbstractPage[ModelType], List[ModelType]]:
        raise NotImplementedError

    @abc.abstractmethod
    async def crud_get_one(
        self,
        request: Request,
        id: ID_TYPE,
        joins: Optional[JoinOptions] = None,
    ) -> ModelType:
        raise NotImplementedError

    @abc.abstractmethod
    async def crud_create_one(
        self,
        request: Request,
        model: CreateSchemaType,
        background_tasks: Optional[BackgroundTasks] = None
    ) -> ModelType:
        raise NotImplementedError

    @abc.abstractmethod
    async def crud_create_many(
        self,
        request: Request,
        models: List[CreateSchemaType],
        background_tasks: Optional[BackgroundTasks] = None
    ) -> List[ModelType]:
        raise NotImplementedError

    @abc.abstractmethod
    async def crud_update_one(
        self,
        request: Request,
        id: ID_TYPE,
        model: UpdateSchemaType,
        background_tasks: Optional[BackgroundTasks] = None
    ) -> ModelType:
        raise NotImplementedError

    @abc.abstractmethod
    async def crud_update_many(
        self,
        request: Request,
        ids: List[ID_TYPE],
        model: List[UpdateSchemaType],
        background_tasks: Optional[BackgroundTasks] = None
    ) -> List[ModelType]:
        raise NotImplementedError

    @abc.abstractmethod
    async def crud_delete_many(
        self,
        request: Request,
        ids: List[ID_TYPE],
        soft_delete: Optional[bool] = False,
        background_tasks: Optional[BackgroundTasks] = None
    ) -> List[ModelType]:
        raise NotImplementedError



from .decorator import crud
from .models import AbstractResponseModel, JoinOptionModel
from .config import BetterCrudGlobalConfig
from .helper import get_feature, get_action, decide_should_paginate
from .types import QuerySortDict
from .pagination import Page, AbstractPage
from .depends import GetQuerySearch, GetQuerySorts, GetQueryJoins, GetQueryLoads
from .generator import crud_generator
from .backend import register_backend
from .exceptions import *

from ._version import __version__

__all__ = [
    "crud",
    "crud_generator",
    "AbstractResponseModel",
    "JoinOptionModel",
    "BetterCrudGlobalConfig",
    "get_feature",
    "get_action",
    "decide_should_paginate",
    "register_backend",
    "QuerySortDict",
    "Page",
    "AbstractPage",
    "GetQuerySearch",
    "GetQuerySorts",
    "GetQueryJoins",
    "GetQueryLoads"
]



from typing import TypeVar, Any
from typing import (
    Any,
    Optional,
    TypeVar,
    Dict,
    Callable,
    Awaitable
)
from fastapi import APIRouter
from .models import CrudOptions
from .types import (
    RoutesModelDict,
    QueryOptionsDict,
    AuthModelDict,
    DtoModelDict,
    SerializeModelDict,
    PathParamDict
)
from .config import BetterCrudGlobalConfig
from .factory import crud_routes_factory
from .service.abstract import AbstractCrudService
from .backend import get_backend

ModelType = TypeVar("ModelType", bound=Any)


def crud_generator(
    router: APIRouter,
    model: ModelType,
    *,
    serialize: SerializeModelDict,
    params: Optional[Dict[str, PathParamDict]] = None,
    routes: Optional[RoutesModelDict] = {},
    dto: DtoModelDict = {},
    auth: Optional[AuthModelDict] = {},
    query: Optional[QueryOptionsDict] = {},
    summary_vars: Optional[Dict] = {},
    feature: Optional[str] = "",
    service_cls: Optional[AbstractCrudService] = None,
    on_before_create: Optional[Callable[..., Awaitable[Any]]] = None,
    on_after_create: Optional[Callable[..., Awaitable[Any]]] = None,
    on_before_update: Optional[Callable[..., Awaitable[Any]]] = None,
    on_after_update: Optional[Callable[..., Awaitable[Any]]] = None,
    on_before_delete: Optional[Callable[..., Awaitable[Any]]] = None,
    on_after_delete: Optional[Callable[..., Awaitable[Any]]] = None
):
    options = CrudOptions(
        feature=feature,
        dto=dto,
        auth=auth,
        params=params,
        serialize=serialize,
        summary_vars=summary_vars,
        routes={**BetterCrudGlobalConfig.routes.model_dump(), **routes},
        query={**BetterCrudGlobalConfig.query.model_dump(), **query}
    )
    backend_name = BetterCrudGlobalConfig.backend_config.backend
    backend_cls = get_backend(backend_name)

    class LocalCrudService(backend_cls[model]):
        def __init__(self):
            super().__init__(model)

        async def on_before_create(self, *args, **kwargs) -> None:
            on_before_create and await on_before_create(*args, **kwargs)

        async def on_after_create(self, *args, **kwargs) -> None:
            on_after_create and await on_after_create(*args, **kwargs)

        async def on_before_update(self, *args, **kwargs) -> None:
            on_before_update and await on_before_update(*args, **kwargs)

        async def on_after_update(self, *args, **kwargs) -> None:
            on_after_update and await on_after_update(*args, **kwargs)

        async def on_before_delete(self, *args, **kwargs) -> None:
            on_before_delete and await on_before_delete(*args, **kwargs)

        async def on_after_delete(self, *args, **kwargs) -> None:
            on_after_delete and await on_after_delete(*args, **kwargs)

    class LocalCrudController:
        def __init__(self):
            self.service = service_cls() if service_cls else LocalCrudService()
    crud_routes_factory(router, LocalCrudController, options)



from typing import Optional, Callable, List, TypeVar, Dict
from fastapi import Request
import json
from fastapi_pagination.api import resolve_params
from fastapi_pagination.bases import AbstractParams
from .types import QuerySortDict
from .models import SerializeModel, RouteOptions, PathParamModel, JoinOptions
from .enums import RoutesEnum

from .config import BetterCrudGlobalConfig
FindType = TypeVar('FindType')


def find(data: List[FindType], fun: Callable[[FindType], bool]) -> FindType:
    for item in data:
        if fun(item):
            return item


def get_feature(request: Request) -> Optional[str]:
    return getattr(request.state, 'feature', None)


def get_action(request: Request) -> Optional[str]:
    return getattr(request.state, 'action', None)


def filter_to_search(filter_str: str) -> Dict:
    filters = filter_str.split(BetterCrudGlobalConfig.delim_config.delim)
    field = filters[0]
    operator = filters[1]
    value = filters[2] if len(filters) == 3 else None
    search = {}
    if operator in ["$isnull", "$notnull"]:
        search = {
            field: {
                operator: True
            }
        }
    else:
        search = {
            field: {
                operator: value
            }
        }
    return search


def parse_query_search(
    search_spec: Optional[Dict] = None,
    ors: Optional[List[str]] = None,
    filters: Optional[List[str]] = None,
    option_filter: Optional[Dict] = None,
    auth_filter: Optional[Dict] = None,
    params_filter: Optional[Dict] = None
):
    search = None
    search_list = []
    if search_spec:
        search_list = [search_spec]
    elif filters and ors:
        if len(filters) == 1 and len(ors) == 1:
            search_list = [{
                "$or": [
                    {
                        **filter_to_search(filters[0])
                    },
                    {
                        **filter_to_search(ors[0])
                    }
                ]
            }]
        else:
            search_list = [{
                "$or": [
                    {
                        "$and": list(map(filter_to_search, filters))
                    },
                    {
                        "$and": list(map(filter_to_search, ors))
                    }
                ]
            }]

    elif filters and len(filters) > 0:
        search_list = list(map(filter_to_search, filters))
    elif ors and len(ors) > 0:
        if len(ors) == 1:
            search_list = [filter_to_search(ors[0])]
        else:
            search_list = [{
                "$or": list(map(filter_to_search, ors))
            }]

    if option_filter:
        search_list.append(option_filter)
    if auth_filter:
        search_list.append(auth_filter)
    if params_filter:
        search_list.append(params_filter)
    if len(search_list) > 0:
        search = {"$and": search_list}
    return search


def parse_query_sort(
    raw_query_sorts: Optional[List[str]] = None,
) -> List[QuerySortDict]:
    sorts = []
    for item in raw_query_sorts:
        if BetterCrudGlobalConfig.delim_config.delim_str not in item:
            raise Exception("invalid query sort")
        field, sort = item.split(
            BetterCrudGlobalConfig.delim_config.delim_str
        )
        sorts.append(QuerySortDict(field=field, sort=sort))
    return sorts


def update_entity_attr(entity, update_value: Dict):
    for key, value in update_value.items():
        if value is not None:
            setattr(entity, key, value)


def decide_should_paginate():
    try:
        params: AbstractParams = resolve_params()
        raw_params = (params.to_raw_params().as_limit_offset())
        return raw_params.limit > 0
    except Exception:
        return False


def get_serialize_model(serialize: SerializeModel, router_name):
    serialize_model = getattr(serialize, router_name, None)
    if serialize_model is None:
        if router_name == RoutesEnum.get_one:
            return get_serialize_model(serialize, RoutesEnum.get_many)
        elif router_name == RoutesEnum.create_many:
            return get_serialize_model(serialize, RoutesEnum.create_one)
        elif router_name == RoutesEnum.update_many:
            return get_serialize_model(serialize, RoutesEnum.update_one)
    return serialize_model or getattr(serialize, "base", None)


class DefaultMap(dict):
    def __missing__(self, key):
        return key


def get_route_summary(route_options: RouteOptions, summary_vars: Dict):
    if not route_options:
        return None
    if not route_options.summary:
        return None
    return route_options.summary.format_map(DefaultMap(**summary_vars))


def get_params_filter(params: Dict[str, PathParamModel], request: Request):
    params_filter = {}
    if params and request.path_params:
        for key, value in params.items():
            if key in request.path_params:
                params_filter[value.field] = request.path_params[key]
    return params_filter


def build_join_option_tree(raw_join_options: JoinOptions):
    if raw_join_options is None:
        return []
    join_options = dict(sorted(raw_join_options.items()))
    nodes = []
    node_dict: Dict[str,] = {}
    for field_key, config in join_options.items():
        segments = field_key.split(".")
        parent_key = ".".join(segments[:-1])
        node_data = {
            "field_key": field_key,
            "config": config,
            "children": []
        }
        if parent_key == "":
            node_dict[field_key] = node_data
            nodes.append(node_dict[field_key])
        else:
            parent_node = node_dict.get(parent_key)
            if parent_node:
                node_dict[field_key] = node_data
                parent_node.get("children").append(node_dict[field_key])
    return nodes



import inspect
from typing import (
    Any,
    Optional,
    Callable,
    List,
    Type,
    TypeVar,
    get_type_hints,
    Dict
)
from fastapi import APIRouter
from pydantic import BaseModel
from .models import (
    CrudOptions,
)
from .types import (
    RoutesModelDict,
    QueryOptionsDict,
    AuthModelDict,
    DtoModelDict,
    SerializeModelDict,
    PathParamDict
)
from .config import BetterCrudGlobalConfig
from .factory import crud_routes_factory
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)
T = TypeVar("T")
CONFIG = TypeVar("CONFIG", bound=CrudOptions)
CRUD_CLASS_KEY = "__crud_class__"
UNBIND_KIND_TYPE = (
    inspect.Parameter.VAR_POSITIONAL,
    inspect.Parameter.VAR_KEYWORD
)


def crud(
    router: APIRouter,
    *,
    serialize: SerializeModelDict,
    params: Optional[Dict[str, PathParamDict]] = None,
    routes: Optional[RoutesModelDict] = {},
    dto: DtoModelDict = {},
    auth: Optional[AuthModelDict] = {},
    query: Optional[QueryOptionsDict] = {},
    summary_vars: Optional[Dict] = {},
    feature: Optional[str] = "",
) -> Callable[[Type[T]], Type[T]]:
    def decorator(cls: Type[T]) -> Type[T]:
        options = CrudOptions(
            feature=feature,
            dto=dto,
            auth=auth,
            params=params,
            serialize=serialize,
            summary_vars=summary_vars,
            routes={**BetterCrudGlobalConfig.routes.model_dump(), **routes},
            query={**BetterCrudGlobalConfig.query.model_dump(), **query}
        )
        _init_cbv(cls)
        return crud_routes_factory(router, cls, options)
    return decorator


def _init_cbv(cls: Type[Any]) -> None:
    if getattr(cls, CRUD_CLASS_KEY, False):
        return
    old_init: Callable[..., Any] = cls.__init__
    old_signature = inspect.signature(old_init)
    old_parameters = list(old_signature.parameters.values())[
        1:]
    new_parameters = [
        x for x in old_parameters if x.kind not in UNBIND_KIND_TYPE
    ]
    dependency_names: List[str] = []
    for name, hint in get_type_hints(cls).items():
        parameter_kwargs = {"default": getattr(cls, name, Ellipsis)}
        dependency_names.append(name)
        new_parameters.append(
            inspect.Parameter(
                name=name,
                kind=inspect.Parameter.KEYWORD_ONLY,
                annotation=hint,
                **parameter_kwargs
            )
        )
    new_signature = old_signature.replace(parameters=new_parameters)

    def new_init(self: Any, *args: Any, **kwargs: Any) -> None:
        for dep_name in dependency_names:
            dep_value = kwargs.pop(dep_name)
            setattr(self, dep_name, dep_value)
        old_init(self, *args, **kwargs)

    setattr(cls, "__signature__", new_signature)
    setattr(cls, "__init__", new_init)
    setattr(cls, CRUD_CLASS_KEY, True)



from pathlib import Path

import setuptools

VERSION = "0.0.5"

NAME = "better_crud"

INSTALL_REQUIRES = [
    "fastapi>=0.111.0",
    "sqlalchemy>=2.0.30",
    "fastapi_pagination>=0.12.24",
    "pydantic>=2.7.3"
]


setuptools.setup(
    name=NAME,
    keywords=["fastapi", "crud", "async", "sqlalchemy", "pydantic"],
    version=VERSION,
    description="A better CRUD library for FastAPI.",
    url="https://github.com/bigrivi/better_crud",
    project_urls={
        "Source Code": "https://github.com/bigrivi/better_crud",
    },
    author="bigrivi",
    author_email="sunjianghong@gmail.com",
    license="Apache License 2.0",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
        "Framework :: FastAPI",
        "Typing :: Typed",
    ],
    python_requires=">=3.9",
    install_requires=INSTALL_REQUIRES,
    packages=["better_crud", "better_crud/service",
              "better_crud/service/sqlalchemy"],
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
)
